# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.packagecloud`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The packagecloud actor allows you to perform maintenance operations on
repositories hosted by packagecloud.io using their API:

https://packagecloud.io/docs/api

**Required Environment Variables**

:PACKAGECLOUD_ACCOUNT:
  packagecloud account name, i.e. https://packagecloud.io/PACKAGECLOUD_ACCOUNT

:PACKAGECLOUD_TOKEN:
  packagecloud API Token
"""

import datetime
import logging
import os
import re
import sys

from tornado import gen

from tornado_rest_client import api

from kingpin.actors import base
from kingpin.actors import exceptions
from kingpin.constants import REQUIRED

log = logging.getLogger(__name__)

__author__ = "Charles McLaughlin <charles@nextdoor.com>"

ACCOUNT = os.getenv("PACKAGECLOUD_ACCOUNT", None)
TOKEN = os.getenv("PACKAGECLOUD_TOKEN", None)


class PackagecloudAPI(api.RestConsumer):

    ENDPOINT = "https://packagecloud.io/api/v1/"
    CONFIG = {
        "attrs": {
            "packages": {
                "path": (
                    "repos/%account%/%repo%/packages.json"
                    "?per_page={}".format(sys.maxsize)
                ),
                "http_methods": {"get": {}},
            },
            "delete": {
                "path": "repos/%account%/%repo%/%distro_version%/%filename%",
                "http_methods": {"delete": {}},
            },
        },
        "auth": {"user": TOKEN, "pass": ""},
    }


class PackagecloudBase(base.BaseActor):
    """Simple packagecloud Abstract Base Object"""

    def __init__(self, *args, **kwargs):
        """Check required environment variables."""
        super(PackagecloudBase, self).__init__(*args, **kwargs)

        if not ACCOUNT:
            raise exceptions.InvalidCredentials(
                'Missing the "PACKAGECLOUD_ACCOUNT" environment variable.'
            )

        if not TOKEN:
            raise exceptions.InvalidCredentials(
                'Missing the "PACKAGECLOUD_TOKEN" environment variable.'
            )

        rest_client = api.RestClient(timeout=120)
        self._packagecloud_client = PackagecloudAPI(client=rest_client)

    @gen.coroutine
    def _get_all_packages(self, repo):
        """Simple method for fetching a dictionary of all packages in a repo

        Args:
            repo: name of the packagecloud repo to fetch from

        Returns:
            A hash of the packages.
        """
        packages = yield self._packagecloud_client.packages(
            token=TOKEN, account=ACCOUNT, repo=repo
        ).http_get()
        raise gen.Return(packages)

    def _get_package_versions(self, name, packages):
        """Find all versions of a given package.

        Args:
            name: name of the package to look for
            packages: hash of all the packages, as returned by the API

        Returns:
            A hash of package versions sorted by creation date
        """
        versions = [
            {
                "created_at": datetime.datetime.strptime(
                    package["created_at"], "%Y-%m-%dT%H:%M:%S.%fZ"
                ),
                "distro_version": package["distro_version"],
                "filename": package["package_html_url"].split("/")[-1],
                "name": package["name"],
            }
            for package in packages
            if package["name"] == name
        ]

        versions.sort(key=lambda x: x.get("created_at"), reverse=False)
        return versions

    def _filter_packages(self, regex, packages):
        """Extracts a list of unique package names to delete

        Args:
            regex: regex of package names to delete
            packages: hash of all the packages, as returned by the API

        Returns:
            A list of unique package names that match the delete pattern.
        """
        pattern = re.compile(regex)
        packages_list_to_delete = {
            package["name"] for package in packages if pattern.match(package["name"])
        }

        self.log.debug(
            "List of packages matching regex (%s): %s"
            % (regex, packages_list_to_delete)
        )

        return packages_list_to_delete

    @gen.coroutine
    def _delete(self, regex, repo, older_than=0, number_to_keep=0):
        """Generic packagecloud delete method, optionally supporting deleting
        old packages by date and/or keeping a certain number of packages.

        Args:
            regex: Regex of packages to delete, e.g. pkg1|pkg2
            repo: name of the packagecloud repo to delete from
            older_than: Delete packages created before this number of seconds
            number_to_keep: Keep at least this number of each package

        Returns:
            A list of the packages that were deleted
        """
        all_packages = yield self._get_all_packages(repo=repo)
        packages_list_to_delete = self._filter_packages(regex, all_packages)
        all_packages_deleted = []

        # Loop through each unique package to delete
        for name in packages_list_to_delete:
            package_versions = self._get_package_versions(name, all_packages)

            # Create a tally of the packages we delete -- usd to give the user
            # a final helpful log statement about the work we did.
            packages_deleted = []

            # Get a total count of the number of versions of this package in
            # the repo -- this variable will then be counted down as we loop,
            # to prevent us from deleting more than 'number_to_keep' packages.
            number_in_repo = len(package_versions)
            self.log.debug("Scanning %s versions (%s)" % (name, number_in_repo))

            for package in package_versions:
                # Safety check -- if there aren't more than the number_to_keep
                # in the repo, then don't bother continuing through the loop
                # for this package. Break out and move to the next name in
                # packages_list_to_delete.
                if number_in_repo <= number_to_keep:
                    self.log.debug(
                        "%s has only %s package versions left, skipping"
                        % (name, number_in_repo)
                    )
                    break

                # If older_than (time in seconds) was supplied, figure out how
                # old the package is. If the package_age (in seconds) is
                # younger than allowed_age (in seconds), then skip to the next
                # package version in the set.
                if older_than:
                    package_age = datetime.datetime.now() - package["created_at"]
                    allowed_age = datetime.timedelta(seconds=older_than)
                    if package_age <= allowed_age:
                        self.log.debug(
                            "%s/%s is only %s old, skipping deletion"
                            % (package["distro_version"], package["name"], package_age)
                        )
                        continue

                # Finally if we got here, then we have enough packages left in
                # the repo, AND (optionally) this package is older than our
                # cutoff age... so delete the package.
                msg = "%s/%s/%s" % (
                    repo,
                    package["distro_version"],
                    package["filename"],
                )
                if self._dry:
                    self.log.info("Would have deleted %s" % msg)
                else:
                    self.log.info("Deleting %s" % msg)

                    yield self._packagecloud_client.delete(
                        token=TOKEN,
                        account=ACCOUNT,
                        repo=repo,
                        distro_version=package["distro_version"],
                        filename=package["filename"],
                    ).http_delete()

                # Decrement list of packages to track how many are left
                number_in_repo = number_in_repo - 1

                # Track *every* package we delete in one big dict -- this is
                # used purely for the unit tests to validate which packages
                # were deleted.
                all_packages_deleted.append(package)

                # Track that this package was deleted -- used in the parent for
                # loop to give the user a final tally of the packages that
                # were kept, and that were deleted.
                packages_deleted.append(package)

            # Print out the packages that were not deleted and left in the repo
            all_files = [
                "%s/%s" % (package["distro_version"], package["filename"])
                for package in package_versions
            ]
            deleted_files = [
                "%s/%s" % (package["distro_version"], package["filename"])
                for package in packages_deleted
            ]
            files_left = list(set(all_files) - set(deleted_files))
            self.log.debug("%s remaining packages: %s" % (name, files_left))

        raise gen.Return(all_packages_deleted)


class Delete(PackagecloudBase):
    """Deletes packages from a PackageCloud repo.

    Searches for packages that match the `packages_to_delete` regex pattern and
    deletes them.  If `number_to_keep` is set, we always at least this number
    of versions of the given package intact in the repo.  Also if
    `number_to_keep` is set, the older versions of a package (based on upload
    time) packages will be deleted first effectively leaving newer packages
    in the repo.

    **Options**

    :number_to_keep:
      Keep at least this number of each package
      (defaults to *0*)

    :packages_to_delete:
      Regex of packages to delete, e.g. pkg1|pkg2

    :repo:
      Which packagecloud repo to delete from

    **Examples**

    .. code-block:: json

      { "desc": "packagecloud Delete example",
        "actor": "packagecloud.Delete",
        "options": {
          "number_to_keep": 10,
          "packages_to_delete": "deleteme",
          "repo": "test"
        }
      }

    """

    all_options = {
        "number_to_keep": (int, 0, "Keep at least this number of each package"),
        "packages_to_delete": (
            str,
            REQUIRED,
            "Regex of packages to delete, e.g. pkg1|pkg2",
        ),
        "repo": (str, REQUIRED, "Which packagecloud repo to delete from"),
    }

    desc = "Deleting {repo}/{packages_to_delete} (keeping {number_to_keep})"

    def __init__(self, *args, **kwargs):
        """Check required environment variables."""
        super(Delete, self).__init__(*args, **kwargs)

        try:
            re.compile(self.option("packages_to_delete"))
        except re.error:
            raise exceptions.InvalidOptions("packages_to_delete is an invalid regex")

    @gen.coroutine
    def _execute(self):
        """Deletes all packages that match the `packages_to_delete` pattern"""
        yield self._delete(
            regex=self.option("packages_to_delete"),
            number_to_keep=self.option("number_to_keep"),
            repo=self.option("repo"),
        )


class DeleteByDate(PackagecloudBase):
    """Deletes packages from a PackageCloud repo older than X.

    Adds additional functionality to the `Delete` class with a `older_than`
    option.  Only packages older than that number of seconds will be deleted.

    **Options**

    :number_to_keep:
      Keep at least this number of each package
      (defaults to *0*)

    :older_than:
      Delete packages created before this number of seconds

    :packages_to_delete:
      Regex of packages to delete, e.g. pkg1|pkg2

    :repo:
      Which packagecloud repo to delete from

    **Examples**

    .. code-block:: json

      { "desc": "packagecloud DeleteByDate example",
        "actor": "packagecloud.DeleteByDate",
        "options": {
          "number_to_keep": 10,
          "older_than": 600,
          "packages_to_delete": "deleteme",
          "repo": "test"
        }
      }

    """

    all_options = {
        "number_to_keep": (int, 0, "Keep at least this number of each package"),
        "older_than": (
            int,
            REQUIRED,
            "Delete packages created before this number of seconds",
        ),
        "packages_to_delete": (
            str,
            REQUIRED,
            "Regex of packages to delete, e.g. pkg1|pkg2",
        ),
        "repo": (str, REQUIRED, "Which packagecloud repo to delete from"),
    }

    desc = "Deleting {repo}/{packages_to_delete} older than {older_than}"

    @gen.coroutine
    def _execute(self):
        yield self._delete(
            regex=self.option("packages_to_delete"),
            number_to_keep=self.option("number_to_keep"),
            older_than=self.option("older_than"),
            repo=self.option("repo"),
        )


class WaitForPackage(PackagecloudBase):
    """Searches for a package that matches `name` and `version` until found or
    a timeout occurs.

    **Options**

    :name:
      Name of the package to search for as a regex

    :version:
      Version of the package to search for as a regex

    :repo:
      Which packagecloud repo to delete from

    :sleep:
      Number of seconds to sleep for between each search

    **Examples**

    .. code-block:: json

      { "desc": "packagecloud WaitForPackage example",
        "actor": "packagecloud.WaitForPackage",
        "options": {
          "name": "findme",
          "version": "0.1",
          "repo": "test",
          "sleep": 10,
        }
      }

    """

    all_options = {
        "name": (str, REQUIRED, "Name of the package to search for as a regex"),
        "version": (str, ".*", "Version of the package to search for as a regex"),
        "repo": (str, REQUIRED, "Which packagecloud repo to search"),
        "sleep": (int, 10, "Number of seconds to sleep for between each search"),
    }

    desc = "Waiting for {repo}/{name}@{version} (up to {sleep}s)"

    def __init__(self, *args, **kwargs):
        """Check required environment variables."""
        super(WaitForPackage, self).__init__(*args, **kwargs)

        try:
            re.compile(self.option("name"))
        except re.error:
            raise exceptions.InvalidOptions("name is an invalid regex")

        try:
            re.compile(self.option("version"))
        except re.error:
            raise exceptions.InvalidOptions("version is an invalid regex")

    @gen.coroutine
    def _search(self, repo, name, version):
        """Searches for a given package until found or a timeout occurs.

        Args:
            repo: name of the repo to search
            name: Name of the package to search for as a regex
            version: Version of the package to search for as a regex

        Returns:
            A list of the packages that were found
        """

        all_packages = yield self._get_all_packages(repo=repo)
        self.log.debug("Found all packages: %s" % all_packages)

        name_pattern = re.compile(name)
        version_pattern = re.compile(version)

        matched_packages = [
            p
            for p in all_packages
            if name_pattern.match(p["name"]) and version_pattern.match(p["version"])
        ]

        raise gen.Return(matched_packages)

    @gen.coroutine
    def _execute(self):
        """Execute method for the WaitForPackage actor"""
        while True:
            self.log.info(
                "Searching for %s %s..." % (self.option("name"), self.option("version"))
            )

            matched_packages = yield self._search(
                repo=self.option("repo"),
                name=self.option("name"),
                version=self.option("version"),
            )

            if len(matched_packages) > 0:
                self.log.info("Found it!")
                raise gen.Return()

            self.log.debug("Not found, sleeping for (%s)" % self.option("sleep"))
            yield gen.sleep(self.option("sleep"))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

import jsonschema

from kingpin import exceptions

__author__ = "Matt Wise <matt@nextdoor.com>"


ACTOR_SCHEMA = {
    "type": ["object"],
    "required": ["actor"],
    "additionalProperties": False,
    "properties": {
        "desc": {"type": "string"},
        "actor": {"type": "string"},
        "options": {
            # 'Options' are Actor specific. However, we validate some internal
            # option types here, if they are supplied.
            "type": "object",
            # Since options are actor specific, ignore unexpected options.
            "additionalProperties": True,
            # Internally expected properties
            "properties": {
                # 'acts' are lists of actors that should be instantiated. Each
                # object should look like this actual schema (with a desc,
                # actor and option key)
                "acts": {
                    "type": "array",
                    # This is a reference to 'self' ... in other words,
                    # this array can only contain other SCHEMA_1_0 type
                    # objets.
                    "items": {"$ref": "#"},
                },
            },
        },
        # Not required. In code, will default to False.
        "warn_on_failure": {"type": ["boolean", "string"]},
        # Not required. In code, will default to <actor>.default_timeout
        "timeout": {"type": ["string", "integer", "number"]},
        # Optional conditional to indicate to skip this actor.
        "condition": {"type": ["boolean", "string"], "default": True},
    },
}

SCHEMA_1_0 = {
    "definitions": {"actor": ACTOR_SCHEMA},
    "anyOf": [
        {"$ref": "#/definitions/actor"},
        {"type": "array", "items": {"$ref": "#/definitions/actor"}},
    ],
}


def validate(config):
    """Validates the JSON against our schemas.

    TODO: Support multiple schema versions

    Args:
        config: Dictionary of parsed JSON

    Returns:
        None: if all is well

    Raises:
        Execption if something went wrong.
    """
    try:
        return jsonschema.validate(config, SCHEMA_1_0)
    except jsonschema.exceptions.ValidationError as e:
        raise exceptions.InvalidScript(e)



"""Tests for the actors.packagecloud package"""

import datetime
import mock

from tornado import testing

from kingpin.actors import exceptions
from kingpin.actors import packagecloud
from kingpin.actors.test.helper import mock_tornado, tornado_value

__author__ = "Charles McLaughlin <charles@nextdoor.com>"

ALL_PACKAGES_MOCK_RESPONSE = [
    {
        "name": "unittest",
        "uploader_name": "unittest",
        "created_at": "2015-07-07T20:27:18.000Z",
        "distro_version": "ubuntu/trusty",
        "filename": "unittest_0.2-1_all.deb",
        "epoch": 0,
        "version": "0.2",
        "private": True,
        "release": "1",
        "package_url": (
            "/api/v1/repos/unittest/test/package/deb/ubuntu/"
            "trusty/unittest/all/0.2-1.json"
        ),
        "type": "deb",
        "package_html_url": (
            "/unittest/test/packages/ubuntu/trusty/" "unittest_0.2-1_all.deb"
        ),
        "repository_html_url": "/unittest/test",
    },
    {
        "name": "unittest",
        "uploader_name": "unittest",
        "created_at": "2014-07-07T20:27:18.000Z",
        "distro_version": "ubuntu/trusty",
        "filename": "unittest_0.1-1_all.deb",
        "epoch": 0,
        "version": "0.1",
        "private": True,
        "release": "1",
        "package_url": (
            "/api/v1/repos/unittest/test/package/deb/ubuntu/"
            "trusty/unittest/all/0.1-1.json"
        ),
        "type": "deb",
        "package_html_url": (
            "/unittest/test/packages/ubuntu/trusty/" "unittest_0.1-1_all.deb"
        ),
        "repository_html_url": "/unittest/test",
    },
]


def _get_older_than():
    """Method for getting an `older_than` value which is used in a few
    tests below.

    The tests specifically look for the second package in
    ALL_PACKAGES_MOCK_RESPONSE, as it's older than the first"""

    created_at = datetime.datetime.strptime(
        ALL_PACKAGES_MOCK_RESPONSE[0]["created_at"], "%Y-%m-%dT%H:%M:%S.%fZ"
    )
    older_than = (
        datetime.datetime.now() - created_at + datetime.timedelta(seconds=600)
    )  # Adding 5m in case the tests run long
    return older_than


class TestPackagecloudBase(testing.AsyncTestCase):
    """Unit tests for the packagecloud Base actor."""

    def setUp(self, *args, **kwargs):
        super(TestPackagecloudBase, self).setUp(*args, **kwargs)
        packagecloud.TOKEN = "Unittest"
        packagecloud.ACCOUNT = "Unittest"
        self.maxDiff = None

    @testing.gen_test
    def test_init_missing_token(self):
        # Un-set the token and make sure the init fails
        packagecloud.TOKEN = None
        with self.assertRaises(exceptions.InvalidCredentials):
            packagecloud.PackagecloudBase("Unit Test Action", {})

    @testing.gen_test
    def test_init_missing_account(self):
        # Un-set the account and make sure the init fails
        packagecloud.ACCOUNT = None
        with self.assertRaises(exceptions.InvalidCredentials):
            packagecloud.PackagecloudBase("Unit Test Action", {})

    @testing.gen_test
    def test_get_all_packages(self):
        actor = packagecloud.PackagecloudBase("Unit test action", {})
        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE
        )
        all_packages = yield actor._get_all_packages(repo="unittest")
        self.assertEqual(all_packages, ALL_PACKAGES_MOCK_RESPONSE)

    @testing.gen_test
    def test_get_package_versions(self):
        actor = packagecloud.PackagecloudBase("Unit test action", {})
        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE
        )
        packages = yield actor._get_all_packages(repo="unittest")
        versions = actor._get_package_versions(name="unittest", packages=packages)

        self.assertEqual(
            versions,
            [
                {
                    "created_at": datetime.datetime(2014, 7, 7, 20, 27, 18),
                    "name": "unittest",
                    "distro_version": "ubuntu/trusty",
                    "filename": "unittest_0.1-1_all.deb",
                },
                {
                    "created_at": datetime.datetime(2015, 7, 7, 20, 27, 18),
                    "name": "unittest",
                    "distro_version": "ubuntu/trusty",
                    "filename": "unittest_0.2-1_all.deb",
                },
            ],
        )

    @testing.gen_test
    def test_filter_packages(self):
        actor = packagecloud.PackagecloudBase("Unit test action", {})
        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE
        )
        packages = yield actor._get_all_packages(repo="unittest")
        packages_list_to_delete = actor._filter_packages(
            regex="unittest", packages=packages
        )
        self.assertEqual(packages_list_to_delete, set(["unittest"]))

    @testing.gen_test
    def test_delete(self):
        actor = packagecloud.PackagecloudBase("Unit test action", {})
        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE
        )
        actor._packagecloud_client.delete().http_delete = mock_tornado({})

        deleted_packages = yield actor._delete(regex="unittest", repo="unittest")

        self.assertEqual(
            deleted_packages,
            [
                {
                    "created_at": datetime.datetime(2014, 7, 7, 20, 27, 18),
                    "name": "unittest",
                    "distro_version": "ubuntu/trusty",
                    "filename": "unittest_0.1-1_all.deb",
                },
                {
                    "created_at": datetime.datetime(2015, 7, 7, 20, 27, 18),
                    "name": "unittest",
                    "distro_version": "ubuntu/trusty",
                    "filename": "unittest_0.2-1_all.deb",
                },
            ],
        )

    @testing.gen_test
    def test_delete_dry(self):
        actor = packagecloud.PackagecloudBase("Unit test action", {})
        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE
        )
        actor._packagecloud_client.delete().http_delete = mock_tornado({})
        actor._dry = True

        deleted_packages = yield actor._delete(regex="unittest", repo="unittest")

        self.assertEqual(
            deleted_packages,
            [
                {
                    "created_at": datetime.datetime(2014, 7, 7, 20, 27, 18),
                    "name": "unittest",
                    "distro_version": "ubuntu/trusty",
                    "filename": "unittest_0.1-1_all.deb",
                },
                {
                    "created_at": datetime.datetime(2015, 7, 7, 20, 27, 18),
                    "name": "unittest",
                    "distro_version": "ubuntu/trusty",
                    "filename": "unittest_0.2-1_all.deb",
                },
            ],
        )

    @testing.gen_test
    def test_delete_keep_one(self):
        actor = packagecloud.PackagecloudBase("Unit test action", {})
        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE
        )
        actor._packagecloud_client.delete().http_delete = mock_tornado({})

        deleted_packages = yield actor._delete(
            regex="unittest", repo="unittest", number_to_keep=1
        )

        self.assertEqual(
            deleted_packages,
            [
                {
                    "created_at": datetime.datetime(2014, 7, 7, 20, 27, 18),
                    "name": "unittest",
                    "distro_version": "ubuntu/trusty",
                    "filename": "unittest_0.1-1_all.deb",
                }
            ],
        )

    @testing.gen_test
    def test_delete_older_than(self):
        actor = packagecloud.PackagecloudBase("Unit test action", {})
        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE
        )
        actor._packagecloud_client.delete().http_delete = mock_tornado({})

        older_than = _get_older_than()

        deleted_packages = yield actor._delete(
            regex="unittest", repo="unittest", older_than=older_than.total_seconds()
        )

        self.assertEqual(
            deleted_packages,
            [
                {
                    "created_at": datetime.datetime(2014, 7, 7, 20, 27, 18),
                    "name": "unittest",
                    "distro_version": "ubuntu/trusty",
                    "filename": "unittest_0.1-1_all.deb",
                }
            ],
        )


class TestDelete(testing.AsyncTestCase):
    """Unit tests for the packagecloud Delete actor."""

    def setUp(self, *args, **kwargs):
        super(TestDelete, self).setUp(*args, **kwargs)
        packagecloud.TOKEN = "Unittest"
        packagecloud.ACCOUNT = "Unittest"
        self.maxDiff = None

    @testing.gen_test
    def test_bad_regex_packages_to_delete(self):
        with self.assertRaises(exceptions.InvalidOptions):
            packagecloud.Delete(
                "Unit test action", {"packages_to_delete": "[", "repo": "unittest"}
            )

    @testing.gen_test
    def test_execute(self):
        actor = packagecloud.Delete(
            "Unit test action", {"packages_to_delete": "unittest", "repo": "unittest"}
        )
        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE
        )
        actor._packagecloud_client.delete().http_delete = mock_tornado({})

        deleted_packages = yield actor._execute()

        self.assertEqual(deleted_packages, None)


class TestDeleteByDate(testing.AsyncTestCase):
    """Unit tests for the packagecloud DeleteByDate actor."""

    def setUp(self, *args, **kwargs):
        super(TestDeleteByDate, self).setUp(*args, **kwargs)
        packagecloud.TOKEN = "Unittest"
        packagecloud.ACCOUNT = "Unittest"
        self.maxDiff = None

    @testing.gen_test
    def test_execute(self):
        older_than = _get_older_than()
        actor = packagecloud.DeleteByDate(
            "Unit test action",
            {
                "packages_to_delete": "unittest",
                "repo": "unittest",
                "older_than": int(older_than.total_seconds()),
            },
        )

        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE
        )
        actor._packagecloud_client.delete().http_delete = mock_tornado({})

        deleted_packages = yield actor._execute()

        self.assertEqual(deleted_packages, None)


class TestWaitForPackage(testing.AsyncTestCase):
    """Unit tests for the packagecloud WaitForPackage actor."""

    def setUp(self, *args, **kwargs):
        super(TestWaitForPackage, self).setUp(*args, **kwargs)
        packagecloud.TOKEN = "Unittest"
        packagecloud.ACCOUNT = "Unittest"
        self.maxDiff = None

    @testing.gen_test
    def test_bad_regex_name(self):
        with self.assertRaises(exceptions.InvalidOptions):
            packagecloud.WaitForPackage(
                "Unit test action", {"name": "[", "version": "1", "repo": "unittest"}
            )

    @testing.gen_test
    def test_bad_regex_version(self):
        with self.assertRaises(exceptions.InvalidOptions):
            packagecloud.WaitForPackage(
                "Unit test action",
                {"name": "unittest", "version": "[", "repo": "unittest"},
            )

    @testing.gen_test
    def test_execute(self):
        actor = packagecloud.WaitForPackage(
            "Unit test action",
            {"name": "unittest", "repo": "unittest", "version": "0.2"},
        )

        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE
        )
        actor._packagecloud_client.delete().http_delete = mock_tornado({})

        matched_packages = yield actor._execute()

        self.assertEqual(matched_packages, None)

    @testing.gen_test
    def test_execute_with_sleep(self):
        actor = packagecloud.WaitForPackage(
            "Unit test action",
            {"name": "not_found", "repo": "unittest", "version": "0.2", "sleep": 1},
        )

        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE
        )
        actor._search = mock.Mock(
            side_effect=[tornado_value([]), tornado_value(["something"])]
        )
        yield actor._execute()
        self.assertEqual(actor._search.call_count, 2)

    @testing.gen_test
    def test_search(self):
        actor = packagecloud.WaitForPackage(
            "Unit test action",
            {"name": "unittest", "repo": "unittest", "version": "0.2"},
        )
        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE
        )

        matched_packages = yield actor._search(
            repo="unittest", name="unittest", version="0.2"
        )

        self.assertEqual(matched_packages, [ALL_PACKAGES_MOCK_RESPONSE[0]])



"""Tests for the actors.slack package"""

import mock

from tornado import testing

from kingpin.actors import slack
from kingpin.actors import exceptions
from kingpin.actors.test.helper import mock_tornado
import importlib


__author__ = "Matt Wise <matt@nextdoor.com>"


class TestSlackBase(testing.AsyncTestCase):
    """Unit tests for the Slack Base actor."""

    def setUp(self, *args, **kwargs):
        # For most tests, mock out the TOKEN
        super(TestSlackBase, self).setUp(*args, **kwargs)
        slack.TOKEN = "Unittest"

    def test_init(self):
        actor = slack.SlackBase("Unit test action", {})

        # Ensure that the actor._slack_client is configured with a dictionary
        # that contains the token in it.
        slack_client_tokens = actor._slack_client._client._tokens
        self.assertEqual(slack_client_tokens, {"token": "Unittest"})

    def test_init_missing_creds(self):
        # Un-set the token now and make sure the init fails
        slack.TOKEN = None
        with self.assertRaises(exceptions.InvalidCredentials):
            slack.SlackBase("Unit Test Action", {})
        # Reload the slack library to re-get the token
        importlib.reload(slack)

    def test_check_results_with_ok_results(self):
        actor = slack.SlackBase("Unit test action", {})
        results = {
            "ok": True,
            "channel": "C03H4GRDF",
            "ts": "1423092527.000006",
            "message": {
                "text": "Hi, testing!",
                "username": "Kingpin",
                "type": "message",
                "subtype": "bot_message",
                "ts": "1423092527.000006",
            },
        }
        self.assertEqual(None, actor._check_results(results))

    def test_check_results_with_invalid_creds(self):
        actor = slack.SlackBase("Unit test action", {})
        results = {"ok": False, "error": "invalid_auth"}
        with self.assertRaises(exceptions.InvalidCredentials):
            actor._check_results(results)

    def test_check_results_with_unexpected_results(self):
        actor = slack.SlackBase("Unit test action", {})
        results = "got some unexpected result"
        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            actor._check_results(results)


class TestMessage(testing.AsyncTestCase):
    """Unit tests for the Slack Message actor."""

    def setUp(self, *args, **kwargs):
        # For most cases, mock out the TOKEN
        super(TestMessage, self).setUp(*args, **kwargs)
        slack.TOKEN = "Unittest"

        self.actor = slack.Message(
            "Unit test message", {"channel": "#testing", "message": "Unittest"}
        )
        self._slack_mock = mock.MagicMock(name="SlackAPIMock")
        self.actor._slack_client = self._slack_mock

    @testing.gen_test
    def test_execute_dry(self):
        # Mock out the calls to SlackAPI.auth_test().auth_test()
        auth_test_mock = mock.MagicMock(name="auth_test")
        auth_test_mock.http_post.side_effect = mock_tornado({"ok": "true"})
        self._slack_mock.auth_test.return_value = auth_test_mock

        # Ensure we're dry
        self.actor._dry = True
        ret = yield self.actor._execute()
        self.assertEqual(None, ret)

        # Ensure the calls were made to the API
        auth_test_mock.http_post.assert_has_calls([mock.call()])

    @testing.gen_test
    def test_execute(self):
        # Mock out the calls to SlackAPI.chat_postMessage().http_post()
        post_mock = mock.MagicMock()
        post_mock.http_post.side_effect = mock_tornado({"ok": "true"})
        self._slack_mock.chat_postMessage.return_value = post_mock

        ret = yield self.actor._execute()
        self.assertEqual(None, ret)

        # Ensure the calls were made to the API
        post_mock.http_post.assert_has_calls(
            [
                mock.call(
                    username="Kingpin",
                    unfurl_links=True,
                    text="Unittest",
                    unfurl_media=True,
                    parse="none",
                    link_names=1,
                    channel="#testing",
                )
            ]
        )

    @testing.gen_test
    def test_execute_list_rooms(self):
        actor = slack.Message(
            "Unit test message",
            {"channel": ["#testing", "#testing2"], "message": "Unittest"},
        )
        actor._slack_client = self._slack_mock

        # Mock out the calls to SlackAPI.chat_postMessage().http_post()
        post_mock = mock.MagicMock()
        post_mock.http_post.side_effect = mock_tornado({"ok": "true"})
        self._slack_mock.chat_postMessage.return_value = post_mock

        ret = yield actor._execute()
        self.assertEqual(None, ret)

        # Ensure the calls were made to the API
        post_mock.http_post.assert_has_calls(
            [
                mock.call(
                    username="Kingpin",
                    unfurl_links=True,
                    text="Unittest",
                    unfurl_media=True,
                    parse="none",
                    link_names=1,
                    channel="#testing",
                )
            ]
        )

    @testing.gen_test
    def test_execute_csv_rooms(self):
        actor = slack.Message(
            "Unit test message",
            {"channel": "#testing, #testing2", "message": "Unittest"},
        )
        actor._slack_client = self._slack_mock

        # Mock out the calls to SlackAPI.chat_postMessage().http_post()
        post_mock = mock.MagicMock()
        post_mock.http_post.side_effect = mock_tornado({"ok": "true"})
        self._slack_mock.chat_postMessage.return_value = post_mock

        ret = yield actor._execute()
        self.assertEqual(None, ret)

        # Ensure the calls were made to the API
        post_mock.http_post.assert_has_calls(
            [
                mock.call(
                    username="Kingpin",
                    unfurl_links=True,
                    text="Unittest",
                    unfurl_media=True,
                    parse="none",
                    link_names=1,
                    channel="#testing",
                )
            ]
        )



"""Tests for the pingdom actors"""

import mock

from tornado import testing

from kingpin.actors import pingdom
from kingpin.actors import exceptions
from kingpin.actors.test.helper import mock_tornado, tornado_value


__author__ = "Mikhail Simin <mikhail@nextdoor.com>"


class TestPingdomBase(testing.AsyncTestCase):
    def setUp(self, *args, **kwargs):
        super(TestPingdomBase, self).setUp()
        pingdom.TOKEN = "Unittest"
        pingdom.USER = "Unittest"
        pingdom.PASS = "Unittest"

    @testing.gen_test
    def test_check_name(self):
        actor = pingdom.PingdomBase("Unit Test Action", {"name": "lollipop"})

        actor._pingdom_client = mock.Mock()
        actor._pingdom_client.checks().http_get = mock_tornado(
            {"checks": [{"name": "lollipop"}]}
        )

        check = yield actor._get_check()

        self.assertEqual(check["name"], "lollipop")

    @testing.gen_test
    def test_check_name_fail(self):
        actor = pingdom.PingdomBase("Unit Test Action", {"name": "lollipop"})

        actor._pingdom_client = mock.Mock()
        actor._pingdom_client.checks().http_get = mock_tornado({"checks": []})

        with self.assertRaises(exceptions.InvalidOptions):
            yield actor._get_check()


class TestPause(testing.AsyncTestCase):
    def setUp(self, *args, **kwargs):
        super(TestPause, self).setUp()
        pingdom.TOKEN = "Unittest"
        pingdom.USER = "Unittest"
        pingdom.PASS = "Unittest"

    @testing.gen_test
    def test_execute(self):
        actor = pingdom.Pause("Unit Test Action", {"name": "lollipop"})

        actor._pingdom_client = mock.Mock()
        actor._get_check = mock_tornado(
            {"name": "lollipop", "hostname": "http://lollipop.com", "id": "lol"}
        )
        actor._pingdom_client.check().http_put.return_value = tornado_value()

        yield actor._execute()

        self.assertEqual(actor._get_check._call_count, 1)
        actor._pingdom_client.check.assert_called_with(check_id="lol")
        actor._pingdom_client.check().http_put.assert_called_with(paused="true")

    @testing.gen_test
    def test_execute_dry(self):
        actor = pingdom.Pause("Unit Test Action", {"name": "lollipop"}, dry=True)

        actor._pingdom_client = mock.Mock()
        actor._get_check = mock_tornado(
            {"name": "lollipop", "hostname": "http://lollipop.com", "id": "lol"}
        )

        yield actor._execute()

        self.assertEqual(actor._get_check._call_count, 1)
        actor._pingdom_client.check().http_put.assert_not_called()


class TestUnpause(testing.AsyncTestCase):
    def setUp(self, *args, **kwargs):
        super(TestUnpause, self).setUp()
        pingdom.TOKEN = "Unittest"
        pingdom.USER = "Unittest"
        pingdom.PASS = "Unittest"

    @testing.gen_test
    def test_execute(self):
        actor = pingdom.Unpause("Unit Test Action", {"name": "lollipop"})

        actor._pingdom_client = mock.Mock()
        actor._get_check = mock_tornado(
            {"name": "lollipop", "hostname": "http://lollipop.com", "id": "lol"}
        )
        actor._pingdom_client.check().http_put.return_value = tornado_value()

        yield actor._execute()

        self.assertEqual(actor._get_check._call_count, 1)
        actor._pingdom_client.check.assert_called_with(check_id="lol")
        actor._pingdom_client.check().http_put.assert_called_with(paused="false")

    @testing.gen_test
    def test_execute_dry(self):
        actor = pingdom.Unpause("Unit Test Action", {"name": "lollipop"}, dry=True)

        actor._pingdom_client = mock.Mock()
        actor._get_check = mock_tornado(
            {"name": "lollipop", "hostname": "http://lollipop.com", "id": "lol"}
        )

        yield actor._execute()

        self.assertEqual(actor._get_check._call_count, 1)
        actor._pingdom_client.check().http_put.assert_not_called()



import logging

from botocore.exceptions import ClientError
from tornado import testing
import mock

from kingpin.actors import exceptions
from kingpin.actors.aws import s3 as s3_actor
from kingpin.actors.aws import settings
import importlib

log = logging.getLogger(__name__)


class TestBucket(testing.AsyncTestCase):
    def setUp(self):
        super(TestBucket, self).setUp()
        settings.AWS_ACCESS_KEY_ID = "unit-test"
        settings.AWS_SECRET_ACCESS_KEY = "unit-test"
        settings.AWS_SESSION_TOKEN = "unit-test"
        importlib.reload(s3_actor)

        self.actor = s3_actor.Bucket(
            options={
                "name": "test",
                "region": "us-east-1",
                "policy": "examples/aws.s3/amazon_put.json",
                "lifecycle": [
                    {
                        "id": "test",
                        "prefix": "/test",
                        "status": "Enabled",
                        "transition": {
                            "days": 45,
                            "storage_class": "GLACIER",
                        },
                        "noncurrent_version_transition": {
                            "noncurrent_days": 14,
                            "storage_class": "GLACIER",
                        },
                        "expiration": "30",
                    }
                ],
                "logging": {"target": "test_target", "prefix": "/prefix"},
                "public_access_block_configuration": {
                    "block_public_acls": True,
                    "block_public_policy": True,
                    "ignore_public_acls": True,
                    "restrict_public_buckets": True,
                },
                "versioning": False,
                "tags": [],
                "notification_configuration": {"queue_configurations": []},
            }
        )
        self.actor.s3_conn = mock.MagicMock()

    def test_init_with_bogus_logging_config(self):
        with self.assertRaises(exceptions.InvalidOptions):
            s3_actor.Bucket(
                options={"name": "test", "logging": {"invalid_data": "bad_field"}}
            )

    def test_generate_lifecycle_valid_config(self):
        # Validates that the generated config called by the __init__ class is
        # correct based on the actor configuration in the setUp() method above
        self.assertEqual(len(self.actor.lifecycle), 1)

        # Verify that the rule was created with the basic options
        r = self.actor.lifecycle[0]
        self.assertEqual(r["ID"], "test")
        self.assertEqual(r["Filter"]["Prefix"], "/test")
        self.assertEqual(r["Status"], "Enabled")

        # Validate that the string "30" was turned into an Expiration object
        self.assertEqual(r["Expiration"]["Days"], 30)

        # Validate that the transition config was built properly too
        self.assertEqual(r["Transitions"][0]["Days"], 45)

        # Validate that the transition config was built properly too
        self.assertEqual(r["NoncurrentVersionTransitions"][0]["NoncurrentDays"], 14)

    def test_snake_to_camel(self):
        snake = {"i_should_be_taller": {"me_too_man": ["not_me"]}}

        self.assertEqual(
            self.actor._snake_to_camel(snake),
            {"IShouldBeTaller": {"MeTooMan": ["not_me"]}},
        )

    @testing.gen_test
    def test_precache(self):
        self.actor.s3_conn.list_buckets.return_value = {
            "Buckets": [{"Name": "wrong_bucket"}, {"Name": "test"}]
        }

        yield self.actor._precache()
        self.assertTrue(self.actor._bucket_exists)

    @testing.gen_test
    def test_get_state_absent(self):
        ret = yield self.actor._get_state()
        self.assertEqual("absent", ret)

    @testing.gen_test
    def test_get_state_present(self):
        self.actor._bucket_exists = True
        ret = yield self.actor._get_state()
        self.assertEqual("present", ret)

    @testing.gen_test
    def test_set_state_absent(self):
        self.actor._options["state"] = "absent"
        yield self.actor._set_state()
        self.actor.s3_conn.delete_bucket.assert_has_calls([mock.call(Bucket="test")])

    @testing.gen_test
    def test_set_state_present(self):
        self.actor._options["state"] = "present"
        yield self.actor._set_state()
        self.actor.s3_conn.create_bucket.assert_has_calls([mock.call(Bucket="test")])

    @testing.gen_test
    def test_create_bucket(self):
        yield self.actor._create_bucket()
        self.actor.s3_conn.create_bucket.assert_called_with(Bucket="test")

    @testing.gen_test
    def test_create_bucket_new_region(self):
        self.actor._options["region"] = "us-west-1"
        yield self.actor._create_bucket()
        self.actor.s3_conn.create_bucket.assert_called_with(
            Bucket="test", CreateBucketConfiguration={"LocationConstraint": "us-west-1"}
        )

    @testing.gen_test
    def test_verify_can_delete_bucket(self):
        self.actor.s3_conn.list_objects.return_value = {"Contents": [1, 2, 3]}
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._verify_can_delete_bucket()

    @testing.gen_test
    def test_verify_can_delete_bucket_true(self):
        self.actor.s3_conn.list_objects.return_value = {}
        yield self.actor._verify_can_delete_bucket()

    @testing.gen_test
    def test_delete_bucket(self):
        yield self.actor._delete_bucket()
        self.actor.s3_conn.delete_bucket.assert_has_calls([mock.call(Bucket="test")])

    @testing.gen_test
    def test_delete_bucket_409(self):
        self.actor.s3_conn.delete_bucket.side_effect = ClientError(
            {"Error": {"Code": ""}}, "Error"
        )
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._delete_bucket()

    @testing.gen_test
    def test_get_policy(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_policy.return_value = {"Policy": "{}"}
        ret = yield self.actor._get_policy()
        self.assertEqual({}, ret)

    @testing.gen_test
    def test_get_policy_no_bucket(self):
        self.actor._bucket_exists = False
        ret = yield self.actor._get_policy()
        self.assertEqual(ret, None)

    @testing.gen_test
    def test_get_policy_empty(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_policy.side_effect = ClientError(
            {"Error": {"Code": ""}}, "NoSuchBucketPolicy"
        )
        ret = yield self.actor._get_policy()
        self.assertEqual("", ret)

    @testing.gen_test
    def test_get_policy_exc(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_policy.side_effect = ClientError(
            {"Error": {"Code": ""}}, "SomeOtherError"
        )
        with self.assertRaises(ClientError):
            yield self.actor._get_policy()

    @testing.gen_test
    def test_compare_policy(self):
        self.actor._bucket_exists = True
        self.actor.policy = {"test": "policy"}
        self.actor.s3_conn.get_bucket_policy.return_value = {
            "Policy": '{"test": "policy"}'
        }
        ret = yield self.actor._compare_policy()
        self.assertTrue(ret)

    @testing.gen_test
    def test_compare_policy_false(self):
        self.actor._bucket_exists = True
        self.actor.policy = {"test": "bah"}
        self.actor.s3_conn.get_bucket_policy.return_value = {
            "Policy": '{"test": "policy"}'
        }
        ret = yield self.actor._compare_policy()
        self.assertFalse(ret)

    @testing.gen_test
    def test_compare_policy_not_managing(self):
        self.actor._bucket_exists = True
        self.actor.policy = None
        self.actor.s3_conn.get_bucket_policy.return_value = {
            "Policy": '{"test": "policy"}'
        }
        ret = yield self.actor._compare_policy()
        self.assertTrue(ret)

    @testing.gen_test
    def test_set_policy(self):
        self.actor.policy = {}
        yield self.actor._set_policy()
        self.actor.s3_conn.put_bucket_policy.assert_has_calls(
            [mock.call(Bucket="test", Policy="{}")]
        )

    @testing.gen_test
    def test_set_policy_malformed_policy(self):
        self.actor.policy = {}
        self.actor.s3_conn.put_bucket_policy.side_effect = ClientError(
            {"Error": {"Code": ""}}, "MalformedPolicy"
        )
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._set_policy()

        self.actor.s3_conn.put_bucket_policy.assert_has_calls(
            [mock.call(Bucket="test", Policy="{}")]
        )

    @testing.gen_test
    def test_set_policy_client_error(self):
        self.actor.policy = {}
        self.actor.s3_conn.put_bucket_policy.side_effect = ClientError(
            {"Error": {"Code": ""}}, "Some Other Error"
        )
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._set_policy()

        self.actor.s3_conn.put_bucket_policy.assert_has_calls(
            [mock.call(Bucket="test", Policy="{}")]
        )

    @testing.gen_test
    def test_set_policy_delete(self):
        self.actor.policy = ""
        yield self.actor._set_policy()
        self.actor.s3_conn.delete_bucket_policy.assert_has_calls(
            [mock.call(Bucket="test")]
        )

    @testing.gen_test
    def test_get_logging(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_logging.return_value = {
            "LoggingEnabled": {
                "TargetBucket": "Target-Bucket",
                "TargetPrefix": "Target-Prefix",
            }
        }

        ret = yield self.actor._get_logging()
        self.assertEqual(ret, {"target": "Target-Bucket", "prefix": "Target-Prefix"})

    @testing.gen_test
    def test_get_logging_disabled(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_logging.return_value = {}
        ret = yield self.actor._get_logging()
        self.assertEqual(ret, {"target": "", "prefix": ""})

    @testing.gen_test
    def test_get_logging_no_bucket(self):
        self.actor._bucket_exists = False
        ret = yield self.actor._get_logging()
        self.assertEqual(ret, None)

    @testing.gen_test
    def test_set_logging_not_desired(self):
        self.actor._options["logging"] = None
        yield self.actor._set_logging()
        self.assertFalse(self.actor.s3_conn.put_bucket_logging.called)

    @testing.gen_test
    def test_set_logging_disabled(self):
        self.actor._options["logging"] = {"target": "", "prefix": ""}
        yield self.actor._set_logging()
        self.actor.s3_conn.put_bucket_logging.assert_has_calls(
            [mock.call(Bucket="test", BucketLoggingStatus={})]
        )

    @testing.gen_test
    def test_set_logging(self):
        self.actor._options["logging"] = {"target": "tgt", "prefix": "pfx"}
        yield self.actor._set_logging()
        self.actor.s3_conn.put_bucket_logging.assert_has_calls(
            [
                mock.call(
                    Bucket="test",
                    BucketLoggingStatus={
                        "LoggingEnabled": {"TargetPrefix": "pfx", "TargetBucket": "tgt"}
                    },
                )
            ]
        )

    @testing.gen_test
    def test_set_logging_client_error(self):
        self.actor.s3_conn.put_bucket_logging.side_effect = ClientError(
            {"Error": {"Code": ""}}, "Some error"
        )
        with self.assertRaises(s3_actor.InvalidBucketConfig):
            yield self.actor._set_logging()

    @testing.gen_test
    def test_get_versioning(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_versioning.return_value = {"Status": "Enabled"}
        ret = yield self.actor._get_versioning()
        self.assertTrue(ret)

    @testing.gen_test
    def test_get_versioning_no_bucket(self):
        self.actor._bucket_exists = False
        ret = yield self.actor._get_versioning()
        self.assertEqual(None, ret)

    @testing.gen_test
    def test_get_versioning_suspended(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_versioning.return_value = {"Status": "Suspended"}
        ret = yield self.actor._get_versioning()
        self.assertFalse(ret)

    @testing.gen_test
    def test_set_versioning(self):
        self.actor._options["versioning"] = True
        yield self.actor._set_versioning()
        self.actor.s3_conn.put_bucket_versioning.assert_has_calls(
            [mock.call(Bucket="test", VersioningConfiguration={"Status": "Enabled"})]
        )

    @testing.gen_test
    def test_set_versioning_suspended(self):
        self.actor._options["versioning"] = False
        yield self.actor._set_versioning()
        self.actor.s3_conn.put_bucket_versioning.assert_has_calls(
            [mock.call(Bucket="test", VersioningConfiguration={"Status": "Suspended"})]
        )

    @testing.gen_test
    def test_set_versioning_none(self):
        self.actor._options["versioning"] = None
        yield self.actor._set_versioning()
        self.assertFalse(self.actor.s3_conn.put_bucket_versioning.called)

    @testing.gen_test
    def test_get_lifecycle(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_lifecycle_configuration.return_value = {
            "Rules": []
        }
        ret = yield self.actor._get_lifecycle()
        self.assertEqual(ret, [])

    @testing.gen_test
    def test_get_lifecycle_no_bucket(self):
        self.actor._bucket_exists = False
        ret = yield self.actor._get_lifecycle()
        self.assertEqual(None, ret)

    @testing.gen_test
    def test_get_lifecycle_empty(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_lifecycle_configuration.side_effect = ClientError(
            {"Error": {"Code": ""}}, "NoSuchLifecycleConfiguration"
        )
        ret = yield self.actor._get_lifecycle()
        self.assertEqual(ret, [])

    @testing.gen_test
    def test_get_lifecycle_clienterror(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_lifecycle_configuration.side_effect = ClientError(
            {"Error": {"Code": ""}}, "SomeOtherError"
        )
        with self.assertRaises(ClientError):
            yield self.actor._get_lifecycle()

    @testing.gen_test
    def test_compare_lifecycle(self):
        self.actor._bucket_exists = True
        self.actor.lifecycle = [{"test": "test"}]
        self.actor.s3_conn.get_bucket_lifecycle_configuration.return_value = {
            "Rules": [{"test": "test"}]
        }
        ret = yield self.actor._compare_lifecycle()
        self.assertTrue(ret)

    @testing.gen_test
    def test_compare_lifecycle_none(self):
        self.actor.lifecycle = None
        ret = yield self.actor._compare_lifecycle()
        self.assertTrue(ret)

    @testing.gen_test
    def test_compare_lifecycle_diff(self):
        self.actor.lifecycle = [{"test1": "test"}]
        self.actor.s3_conn.get_bucket_lifecycle_configuration.return_value = {
            "Rules": [{"test": "test"}]
        }
        ret = yield self.actor._compare_lifecycle()
        self.assertFalse(ret)

    @testing.gen_test
    def test_set_lifecycle_delete(self):
        self.actor.lifecycle = []
        yield self.actor._set_lifecycle()
        self.actor.s3_conn.delete_bucket_lifecycle.assert_has_calls(
            [mock.call(Bucket="test")]
        )

    @testing.gen_test
    def test_set_lifecycle(self):
        self.actor.lifecycle = [{"test": "test"}]
        yield self.actor._set_lifecycle()
        self.actor.s3_conn.put_bucket_lifecycle_configuration.assert_has_calls(
            [
                mock.call(
                    Bucket="test", LifecycleConfiguration={"Rules": [{"test": "test"}]}
                )
            ]
        )

    @testing.gen_test
    def test_set_lifecycle_client_error(self):
        self.actor.s3_conn.put_bucket_lifecycle_configuration.side_effect = ClientError(
            {"Error": {"Code": ""}}, "Error"
        )
        with self.assertRaises(s3_actor.InvalidBucketConfig):
            yield self.actor._set_lifecycle()

    @testing.gen_test
    def test_get_public_access_block_configuration(self):
        test_cfg = {
            "BlockPublicAcls": True,
            "BlockPublicPolicy": True,
            "IgnorePublicAcls": True,
            "RestrictPublicBuckets": True,
        }

        self.actor._bucket_exists = True
        self.actor.s3_conn.get_public_access_block.return_value = {
            "PublicAccessBlockConfiguration": test_cfg
        }
        ret = yield self.actor._get_public_access_block_configuration()
        self.assertEqual(ret, test_cfg)

    @testing.gen_test
    def test_get_public_access_block_configuration_no_bucket(self):
        self.actor._bucket_exists = False
        ret = yield self.actor._get_public_access_block_configuration()
        self.assertEqual(None, ret)

    @testing.gen_test
    def test_get_public_access_block_configuration_empty(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_public_access_block.side_effect = ClientError(
            {"Error": {}}, "NoSuchPublicAccessBlockConfiguration"
        )
        ret = yield self.actor._get_public_access_block_configuration()
        self.assertEqual(ret, [])

    @testing.gen_test
    def test_get_public_access_block_configuration_clienterror(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_public_access_block.side_effect = ClientError(
            {"Error": {}}, "SomeOtherError"
        )
        with self.assertRaises(ClientError):
            yield self.actor._get_public_access_block_configuration()

    @testing.gen_test
    def test_compare_public_access_block_configuration(self):
        self.actor._bucket_exists = True
        self.actor.access_block = [{"test": "test"}]
        self.actor.s3_conn.get_public_access_block.return_value = {
            "PublicAccessBlockConfiguration": [{"test": "test"}]
        }
        ret = yield self.actor._compare_public_access_block_configuration()
        self.assertTrue(ret)

    @testing.gen_test
    def test_compare_public_access_block_configuration_none(self):
        self.actor.access_block = None
        ret = yield self.actor._compare_public_access_block_configuration()
        self.assertTrue(ret)

    @testing.gen_test
    def test_compare_public_access_block_configuration_diff(self):
        self.actor.access_block = [{"test1": "test"}]
        self.actor.s3_conn.get_public_access_block.return_value = {
            "Rules": [{"test": "test"}]
        }
        ret = yield self.actor._compare_public_access_block_configuration()
        self.assertFalse(ret)

    @testing.gen_test
    def test_set_public_access_block_configuration_delete(self):
        self.actor.access_block = {}
        yield self.actor._set_public_access_block_configuration()
        self.actor.s3_conn.delete_public_access_block.assert_has_calls(
            [mock.call(Bucket="test")]
        )

    @testing.gen_test
    def test_set_public_access_block_configuration(self):
        self.actor.access_block = {"test": "test"}
        yield self.actor._set_public_access_block_configuration()
        self.actor.s3_conn.put_public_access_block.assert_has_calls(
            [mock.call(Bucket="test", PublicAccessBlockConfiguration={"test": "test"})]
        )

    @testing.gen_test
    def test_set_public_access_block_configuration_client_error(self):
        self.actor.s3_conn.put_public_access_block.side_effect = ClientError(
            {"Error": {}}, "Error"
        )
        with self.assertRaises(s3_actor.InvalidBucketConfig):
            yield self.actor._set_public_access_block_configuration()

    @testing.gen_test
    def test_get_tags(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_tagging.return_value = {
            "TagSet": [{"Key": "k1", "Value": "v1"}]
        }
        ret = yield self.actor._get_tags()
        self.assertEqual(ret, [{"key": "k1", "value": "v1"}])

    @testing.gen_test
    def test_get_tags_multiple_tags(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_tagging.return_value = {
            "TagSet": [{"Key": "k1", "Value": "v1"}, {"Key": "k2", "Value": "v2"}]
        }
        ret = yield self.actor._get_tags()
        self.assertEqual(
            ret, [{"key": "k1", "value": "v1"}, {"key": "k2", "value": "v2"}]
        )

    @testing.gen_test
    def test_get_tags_no_bucket(self):
        self.actor._bucket_exists = False
        ret = yield self.actor._get_tags()
        self.assertEqual(None, ret)

    @testing.gen_test
    def test_get_tags_not_managed(self):
        self.actor._options["tags"] = None
        ret = yield self.actor._get_tags()
        self.assertEqual(None, ret)

    @testing.gen_test
    def test_get_tags_empty(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_tagging.side_effect = ClientError(
            {"Error": {"Code": ""}}, "NoSuchTagSet"
        )
        ret = yield self.actor._get_tags()
        self.assertEqual(ret, [])

    @testing.gen_test
    def test_get_tags_exc(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_tagging.side_effect = ClientError(
            {"Error": {"Code": ""}}, "SomeOtherError"
        )
        with self.assertRaises(ClientError):
            yield self.actor._get_tags()

    @testing.gen_test
    def test_compare_tags(self):
        self.actor._bucket_exists = True
        self.actor._options["tags"] = [{"key": "test_key", "value": "test_value"}]
        self.actor.s3_conn.get_bucket_tagging.return_value = {
            "TagSet": [
                {"Key": "test_key", "Value": "test_value"},
            ]
        }
        ret = yield self.actor._compare_tags()
        self.assertTrue(ret)

    @testing.gen_test
    def test_compare_tags_false(self):
        self.actor._bucket_exists = True
        self.actor._options["tags"] = [{"key": "test_key", "value": "test_value"}]
        self.actor.s3_conn.get_bucket_tagging.return_value = {
            "TagSet": [{"Key": "k1", "Value": "v1"}, {"Key": "k2", "Value": "v2"}]
        }
        ret = yield self.actor._compare_tags()
        self.assertFalse(ret)

    @testing.gen_test
    def test_compare_tags_not_managing(self):
        self.actor._bucket_exists = True
        self.actor._options["tags"] = None
        ret = yield self.actor._compare_tags()
        self.assertTrue(ret)

    @testing.gen_test
    def test_set_tags_none(self):
        self.actor._options["tags"] = None
        yield self.actor._set_tags()
        self.assertFalse(self.actor.s3_conn.put_bucket_tagging.called)

    @testing.gen_test
    def test_set_tags(self):
        self.actor._options["tags"] = [{"key": "tag1", "value": "v1"}]
        yield self.actor._set_tags()
        self.actor.s3_conn.put_bucket_tagging.assert_has_calls(
            [
                mock.call(
                    Bucket="test", Tagging={"TagSet": [{"Key": "tag1", "Value": "v1"}]}
                )
            ]
        )

    @testing.gen_test
    def test_snake_to_camelcase_for_notification_configuration(self):
        notification_configuration_snake_case = {
            "queue_configurations": [
                {
                    "queue_arn": "arn:aws:sqs:us-east-1:1234567:test_sqs",
                    "events": ["s3:ObjectCreated:*"],
                }
            ]
        }

        notification_configuration_camel_case = self.actor._snake_to_camel(
            notification_configuration_snake_case
        )

        self.assertEqual(
            notification_configuration_camel_case,
            {
                "QueueConfigurations": [
                    {
                        "QueueArn": "arn:aws:sqs:us-east-1:1234567:test_sqs",
                        "Events": ["s3:ObjectCreated:*"],
                    }
                ]
            },
        )

    @testing.gen_test
    def test_set_notification_configurations_none(self):
        self.actor.notification_configuration = None
        yield self.actor._set_notification_configuration()
        self.assertFalse(
            self.actor.s3_conn.put_bucket_notification_configuration.called
        )

    @testing.gen_test
    def test_set_notification_configurations_with_valid_configs(self):
        self.actor.notification_configuration = {
            "QueueConfigurations": [
                {
                    "QueueArn": "arn:aws:sqs:us-east-1:1234567:test_sqs",
                    "Events": ["s3:ObjectCreated:*"],
                }
            ]
        }
        yield self.actor._set_notification_configuration()
        self.actor.s3_conn.put_bucket_notification_configuration.assert_has_calls(
            [
                mock.call(
                    Bucket="test",
                    NotificationConfiguration={
                        "QueueConfigurations": [
                            {
                                "QueueArn": "arn:aws:sqs:us-east-1:1234567:test_sqs",
                                "Events": ["s3:ObjectCreated:*"],
                            }
                        ]
                    },
                )
            ]
        )

    @testing.gen_test
    def test_set_notification_configurations_with_multiple_configs(self):
        self.actor.notification_configuration = {
            "QueueConfigurations": [
                {
                    "QueueArn": "arn:aws:sqs:us-east-1:1:test1_sqs",
                    "Events": ["s3:ObjectCreated:*"],
                },
                {
                    "QueueArn": "arn:aws:sqs:us-east-1:1:test2_sqs",
                    "Events": ["s3:ObjectCreated:*", "s3:ObjectRemoved:*"],
                },
            ]
        }
        yield self.actor._set_notification_configuration()
        self.actor.s3_conn.put_bucket_notification_configuration.assert_has_calls(
            [
                mock.call(
                    Bucket="test",
                    NotificationConfiguration={
                        "QueueConfigurations": [
                            {
                                "QueueArn": "arn:aws:sqs:us-east-1:1:test1_sqs",
                                "Events": ["s3:ObjectCreated:*"],
                            },
                            {
                                "QueueArn": "arn:aws:sqs:us-east-1:1:test2_sqs",
                                "Events": ["s3:ObjectCreated:*", "s3:ObjectRemoved:*"],
                            },
                        ]
                    },
                )
            ]
        )

    @testing.gen_test
    def test_set_notification_configurations_no_configs(self):
        self.actor.notification_configuration = {"QueueConfigurations": []}
        yield self.actor._set_notification_configuration()
        self.actor.s3_conn.put_bucket_notification_configuration.assert_has_calls([])

    @testing.gen_test
    def test_set_notification_configurations_empty_queue_configs(self):
        self.actor.notification_configuration = {}
        yield self.actor._set_notification_configuration()
        self.actor.s3_conn.put_bucket_notification_configuration.assert_has_calls([])

    @testing.gen_test
    def test_get_notification_configuration_with_existing_queueconfigs(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_notification_configuration.return_value = {
            "QueueConfigurations": [
                {"QueueArn": "arn:aws:sqs", "Events": ["s3:ObjectCreated:*"]}
            ]
        }
        ret = yield self.actor._get_notification_configuration()
        self.assertEqual(type(ret), dict)
        self.assertEqual(
            ret,
            {
                "QueueConfigurations": [
                    {"QueueArn": "arn:aws:sqs", "Events": ["s3:ObjectCreated:*"]}
                ]
            },
        )

    @testing.gen_test
    def test_get_notification_configuration_no_bucket(self):
        self.actor._bucket_exists = False
        ret = yield self.actor._get_notification_configuration()
        self.assertEqual(ret, None)

    @testing.gen_test
    def test_get_notification_configuration_with_no_config(self):
        self.actor._bucket_exists = True
        self.actor.notification_configuration = None
        ret = yield self.actor._get_notification_configuration()
        self.assertEqual(ret, None)

    @testing.gen_test
    def test_compare_notification_configuration_with_new_config(self):
        self.actor.notification_configuration = {
            "QueueConfigurations": [
                {
                    "QueueArn": "arn:aws:sqs:us-east-1:1234567:test_sqs",
                    "Events": ["s3:ObjectCreated:*"],
                }
            ]
        }
        self.actor.s3_conn.get_bucket_notification_configuration.return_value = {}
        ret = yield self.actor._compare_notification_configuration()
        self.assertFalse(ret)

    @testing.gen_test
    def test_compare_notification_configuration_with_no_updated_config(self):
        self.actor._bucket_exists = True
        self.actor.notification_configuration = {
            "QueueConfigurations": [
                {
                    "QueueArn": "arn:aws:sqs:us-east-1:1234567:test_sqs",
                    "Events": ["s3:ObjectCreated:*"],
                }
            ]
        }
        self.actor.s3_conn.get_bucket_notification_configuration.return_value = {
            "QueueConfigurations": [
                {
                    "QueueArn": "arn:aws:sqs:us-east-1:1234567:test_sqs",
                    "Events": ["s3:ObjectCreated:*"],
                }
            ]
        }
        ret = yield self.actor._compare_notification_configuration()
        self.assertTrue(ret)

    @testing.gen_test
    def test_compare_notification_configuration_with_no_config(self):
        self.actor.notification_configuration = None
        self.actor.s3_conn.get_bucket_notification_configuration.return_value = {}
        ret = yield self.actor._compare_notification_configuration()
        self.assertTrue(ret)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.aws.settings`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Common settings used by many of the `kingpin.actors.aws` modules.
"""


import os

__author__ = "Mikhail Simin <mikhail@nextdoor.com>"

# By default, this means that Boto will make HTTP calls at instantiation time
# to determine whether or not credentials are available from the metadata
# service.
#
# During tests, we mock these out to blank strings to prevent these calls.
AWS_ACCESS_KEY_ID = os.getenv("AWS_ACCESS_KEY_ID", None)
AWS_SECRET_ACCESS_KEY = os.getenv("AWS_SECRET_ACCESS_KEY", None)
AWS_SESSION_TOKEN = os.getenv("AWS_SESSION_TOKEN", None)

# kingpin is pretty fast which can leads to API throttling. You can set you own
# boto3 configuration by using the standard AWS env vars, but in the absence of
# them, we try to set you some sane defaults based on the boto3 documentation
# and our experience with running kingpin as scale.
#
# Docs: https://boto3.amazonaws.com/v1/documentation/api/latest/guide/retries.html
AWS_MAX_ATTEMPTS = int(os.getenv("AWS_MAX_ATTEMPTS", 10))
AWS_RETRY_MODE = os.getenv("AWS_RETRY_MODE", "standard")

# Set to "" (an empty string) to disable.
KINGPIN_CFN_HASH_OUTPUT_KEY = os.getenv("KINGPIN_CFN_HASH_OUTPUT_KEY", "KingpinCfnHash")



"""Tests for the actors.base package."""

import io
import json
import os
import logging
from importlib import reload

from tornado import gen
from tornado import httpclient
from tornado import simple_httpclient
from tornado import testing
import mock

# Unusual placement -- but we override the environment so that we can test that
# the urllib debugger works.
os.environ["URLLIB_DEBUG"] = "1"

from kingpin import utils
from kingpin.actors import base

reload(base)
from kingpin.actors import exceptions
from kingpin.actors.test.helper import mock_tornado
from kingpin.constants import REQUIRED, STATE


__author__ = "Matt Wise <matt@nextdoor.com>"


class FakeHTTPClientClass(object):
    """Fake HTTPClient object for testing"""

    response_value = None

    @gen.coroutine
    def fetch(self, *args, **kwargs):
        self.request = args[0]
        raise gen.Return(self.response_value)


class FakeEnsurableBaseActor(base.EnsurableBaseActor):

    all_options = {
        "name": (str, REQUIRED, "Name of thing"),
        "description": (str, None, "Some description"),
        "unmanaged": (str, None, "some unmanaged option"),
    }

    unmanaged_options = ["unmanaged"]

    @gen.coroutine
    def _precache(self):
        # Call our parent class precache.. no real need here other than for
        # unit test coverage.
        yield super(FakeEnsurableBaseActor, self)._precache()

        # These do not match -- so we'll trigger the setters
        self.state = "absent"
        self.name = "Old name"

        # This matches the desired description on purpose.
        self.description = "Some description"

        # Start out with no calls recorded
        self.set_state_called = False
        self.set_name_called = False
        self.set_description_called = False

        # Make it easy to check that this was called
        self._precache_called = True

    @gen.coroutine
    def _set_state(self):
        self.state = True
        self.set_state_called = True

    @gen.coroutine
    def _get_state(self):
        raise gen.Return(self.state)

    @gen.coroutine
    def _set_name(self):
        self.name = self.option("name")
        self.set_name_called = True

    @gen.coroutine
    def _get_name(self):
        raise gen.Return(self.name)

    @gen.coroutine
    def _compare_name(self):
        exist = yield self._get_name()
        new = self.option("name")
        raise gen.Return(exist == new)

    @gen.coroutine
    def _set_description(self):
        self.set_description_called = True

    @gen.coroutine
    def _get_description(self):
        raise gen.Return(self.description)


class TestBaseActor(testing.AsyncTestCase):
    @gen.coroutine
    def true(self):
        yield utils.tornado_sleep(0.01)
        raise gen.Return(True)

    @gen.coroutine
    def false(self):
        yield utils.tornado_sleep(0.01)
        raise gen.Return(False)

    def setUp(self):
        super(TestBaseActor, self).setUp()

        # Create a BaseActor object
        self.actor = base.BaseActor("Unit Test Action", {})

        # Mock out the actors ._execute() method so that we have something to
        # test
        self.actor._execute = self.true

    @testing.gen_test
    def test_user_defined_desc(self):
        self.assertEqual("Unit Test Action", str(self.actor))

    @testing.gen_test
    def test_default_desc(self):
        self.actor._desc = None
        self.assertEqual("kingpin.actors.base.BaseActor", str(self.actor))

    @testing.gen_test
    def test_timer(self):
        # Create a function and wrap it in our timer
        self.actor._execute = self.true

        # Mock out the logger so we can track it
        self.actor.log = mock.MagicMock()

        # Now call the execute() wrapper that leverages the @timer decorator.
        yield self.actor.execute()

        # Search for a logged message. Don't explicitly set the execution time
        # because some computers and compilers are slow.
        msg = "kingpin.actors.base.BaseActor.execute() execution time"
        msg_is_in_calls = False
        for call in self.actor.log.debug.mock_calls:
            if msg in str(call):
                msg_is_in_calls = True
        self.assertEqual(msg_is_in_calls, True)

    @testing.gen_test
    def test_timeout(self):
        # Create a quick mock.. so we can track whether or not API calls were
        # actually made.
        tracker = mock.MagicMock(name="tracker")

        # Create a function and wrap it in our timeout
        @gen.coroutine
        def _execute():
            tracker.reset_mock()
            yield gen.sleep(0.2)
            tracker.call_me()

        self.actor._execute = _execute

        # Set our timeout to 2s, test should work
        self.actor._timeout = 1
        yield self.actor.timeout(_execute)
        tracker.assert_has_calls([mock.call.call_me()])

        # Now set our timeout to 500ms. Exception should be raised, and the
        # tracker should NOT be called.
        self.actor._timeout = 0.1
        with self.assertRaises(exceptions.ActorTimedOut):
            yield self.actor.timeout(_execute)

        # Set the timeout to 0, which disables it. No exception should be
        # raised
        self.actor._timeout = 0
        yield self.actor.timeout(_execute)
        self.actor_timeout = None
        yield self.actor.timeout(_execute)

    @testing.gen_test
    def test_httplib_debugging(self):
        # Get the logger now and validate that its level was set right
        requests_logger = logging.getLogger("requests.packages.urllib3")
        self.assertEqual(logging.DEBUG, requests_logger.level)

    def test_validate_options(self):
        self.actor.all_options = {"test": (str, REQUIRED, "")}
        self.actor._options = {"a": "b"}
        with self.assertRaises(exceptions.InvalidOptions):
            ret = self.actor._validate_options()

        self.actor.all_options = {"test": (str, REQUIRED, "")}
        self.actor._options = {"test": "b"}
        ret = self.actor._validate_options()
        self.assertEqual(None, ret)

        self.actor.all_options = {"test": (bool, REQUIRED, "")}
        self.actor._options = {"test": "junk_text"}
        with self.assertRaises(exceptions.InvalidOptions):
            self.actor._validate_options()

        self.actor.all_options = {
            "test": (str, REQUIRED, ""),
            "test2": (str, REQUIRED, ""),
        }
        self.actor._options = {"test": "b", "test2": "b"}
        ret = self.actor._validate_options()
        self.assertEqual(None, ret)

        # The STATE type requires either 'present' or 'absent' to be passed in.
        self.actor.all_options = {"test": (STATE, REQUIRED, "")}
        self.actor._options = {"test": "present"}
        ret = self.actor._validate_options()
        self.assertEqual(None, ret)

        self.actor._options = {"test": "absent"}
        ret = self.actor._validate_options()
        self.assertEqual(None, ret)

        with self.assertRaises(exceptions.InvalidOptions):
            self.actor._options = {"test": "abse"}
            ret = self.actor._validate_options()

    def test_validation_issues(self):
        self.actor.all_options = {
            "needed": (str, REQUIRED, ""),
            "optional": (str, "", ""),
        }

        # Requirement not satisfied
        self.actor._options = {"optional": "b"}
        with self.assertRaises(exceptions.InvalidOptions):
            self.actor._validate_options()

        # Invalid option type:
        self.actor._options = {"needed": 1, "optional": "b"}
        with self.assertRaises(exceptions.InvalidOptions):
            self.actor._validate_options()

        # Unexpected option passed
        self.actor._options = {"needed": "a", "unexpected": "b"}
        # Should work w/out raising an exception.
        self.actor._validate_options()

    def test_validate_defaults(self):
        # Default is not a permitted type
        self.actor.all_options = {"name": (str, False, "String!")}
        self.actor._setup_defaults()
        with self.assertRaises(exceptions.InvalidOptions):
            self.actor._validate_options()

    @testing.gen_test
    def test_option(self):
        self.actor._options["foo"] = "bar"
        opt = self.actor.option("foo")
        self.assertEqual(opt, "bar")

    def test_readfile(self):
        with self.assertRaises(exceptions.InvalidOptions):
            self.actor.readfile("notfound")

        open_patcher = mock.patch("%s.open" % self.actor.__module__, create=True)
        with open_patcher as mock_open:
            self.actor.readfile("somefile")
            self.assertEqual(mock_open.call_count, 1)
            # using __enter__ here because it's opened as a context manager.
            self.assertEqual(mock_open().__enter__().read.call_count, 1)

    @testing.gen_test
    def test_execute(self):
        res = yield self.actor.execute()
        self.assertEqual(res, True)

    def test_str2bool(self):
        self.assertEqual(True, self.actor.str2bool("true"))
        self.assertEqual(True, self.actor.str2bool("junk text"))
        self.assertEqual(True, self.actor.str2bool("1"))
        self.assertEqual(True, self.actor.str2bool(True))
        self.assertEqual(False, self.actor.str2bool("false"))
        self.assertEqual(False, self.actor.str2bool("0"))
        self.assertEqual(False, self.actor.str2bool(False))

    def test_str2bool_strict(self):
        self.assertEqual(True, self.actor.str2bool("true"))
        self.assertEqual(False, self.actor.str2bool(False))
        with self.assertRaises(exceptions.InvalidOptions):
            self.actor.str2bool("Junk", strict=True)

    @testing.gen_test
    def test_check_condition(self):
        conditions = {
            "FOobar": True,
            "True": True,
            "TRUE": True,
            1: True,
            "1": True,
            0: False,
            "0": False,
            "False": False,
            "FALSE": False,
        }
        for value, should_execute in list(conditions.items()):
            self.actor._condition = value
            self.actor._execute = mock_tornado()
            yield self.actor.execute()
            str_value = json.dumps(value)
            if should_execute:
                self.assertEqual(
                    self.actor._execute._call_count,
                    1,
                    "Value `%s` should allow actor execution" % str_value,
                )
            else:
                self.assertEqual(
                    self.actor._execute._call_count,
                    0,
                    "Value `%s` should not allow actor execution" % str_value,
                )

    @testing.gen_test
    def test_execute_fail(self):
        self.actor._execute = self.false
        res = yield self.actor.execute()
        self.assertEqual(res, False)

    @testing.gen_test
    def test_execute_catches_expected_exception(self):
        @gen.coroutine
        def raise_exc():
            raise exceptions.ActorException("Test")

        self.actor._execute = raise_exc
        with self.assertRaises(exceptions.ActorException):
            yield self.actor.execute()

    @testing.gen_test
    def test_execute_catches_unexpected_exception(self):
        @gen.coroutine
        def raise_exc():
            raise Exception("Test")

        self.actor._execute = raise_exc
        with self.assertRaises(exceptions.ActorException):
            yield self.actor.execute()

    @testing.gen_test
    def test_execute_with_warn_on_failure(self):
        @gen.coroutine
        def raise_exc():
            raise exceptions.RecoverableActorFailure("should just warn")

        self.actor._execute = raise_exc

        # First test, should raise an exc...
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor.execute()

        # Second test, turn on 'warn_on_failure'
        self.actor._warn_on_failure = True
        res = yield self.actor.execute()
        self.assertEqual(res, None)

    def test_fill_in_contexts_desc(self):
        base.BaseActor.all_options = {"test_opt": (str, REQUIRED, "Test option")}

        self.actor = base.BaseActor(
            desc="Unit Test Action - {NAME}",
            options={"test_opt": "Foo bar"},
            condition="{NAME}",
            init_context={"NAME": "TEST"},
        )
        self.assertEqual("Unit Test Action - TEST", self.actor._desc)
        self.assertEqual("TEST", self.actor._condition)

        with self.assertRaises(exceptions.InvalidOptions):
            self.actor = base.BaseActor(
                desc="Unit Test Action",
                options={"test_opt": "Foo {BAZ} bar"},
                init_context={},
            )

        with self.assertRaises(exceptions.InvalidOptions):
            self.actor = base.BaseActor(
                desc="Unit Test Action - {NAME}", options={}, init_context={}
            )

        with self.assertRaises(exceptions.InvalidOptions):
            self.actor = base.BaseActor(
                desc="Unit Test Action",
                options={"test_opt": "Foo bar"},
                condition="{NAME}",
                init_context={},
            )

        # Reset the all options so we dont break other tests
        base.BaseActor.all_options = {}

    def test_fill_in_contexts_options(self):
        base.BaseActor.all_options = {"test_opt": (str, REQUIRED, "Test option")}

        self.actor = base.BaseActor(
            desc="Unit Test Action",
            options={"test_opt": "Foo bar - {NAME}"},
            init_context={"NAME": "TEST"},
        )
        self.assertEqual("Foo bar - TEST", self.actor.option("test_opt"))

        # Reset the all options so we dont break other tests
        base.BaseActor.all_options = {}

    def test_fill_in_contexts_options_escape(self):
        base.BaseActor.all_options = {"test_opt": (str, REQUIRED, "Test option")}

        self.actor = base.BaseActor(
            desc="Unit Test Action",
            options={"test_opt": "Foo bar - \\{NAME\\}"},
            init_context={"NAME": "TEST"},
        )
        self.assertEqual("Foo bar - {NAME}", self.actor.option("test_opt"))

        # Reset the all options so we dont break other tests
        base.BaseActor.all_options = {}


class TestEnsurableBaseActor(testing.AsyncTestCase):
    def setUp(self):
        super(TestEnsurableBaseActor, self).setUp()
        self.actor = FakeEnsurableBaseActor(
            "Unit Test Actor",
            {
                "name": "new name",
                "state": "present",
                "unmanaged": "nothing happens with this",
                "description": "Some description",
            },
        )

    @testing.gen_test
    def test_precache(self):
        yield self.actor._precache()

    @testing.gen_test
    def test_execute(self):
        yield self.actor._execute()

        # Did the precache execute?
        self.assertTrue(self.actor._precache_called)

        # First test -- the description should have matched, so
        # we should not have called self._set_description().
        self.assertFalse(self.actor.set_description_called)

        # We _should_ have called the setters for the state, and
        # for the name.
        self.assertTrue(self.actor.set_state_called)
        self.assertTrue(self.actor.set_name_called)

    @testing.gen_test
    def test_execute_absent(self):
        self.actor._options["state"] = "absent"
        yield self.actor._execute()

        # Make sure that the set_name and set_state were NOT called
        self.assertFalse(self.actor.set_state_called)
        self.assertFalse(self.actor.set_name_called)

    @testing.gen_test
    def test_gather_methods_throws_exception(self):
        # Mock out the set_name method by replacing it with an attribute
        self.actor._set_name = False
        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            self.actor._gather_methods()


class TestHTTPBaseActor(testing.AsyncTestCase):
    def setUp(self):
        super(TestHTTPBaseActor, self).setUp()
        self.actor = base.HTTPBaseActor("Unit Test Action", {})

    @testing.gen_test
    def test_get_http_client(self):
        ret = self.actor._get_http_client()
        self.assertEqual(simple_httpclient.SimpleAsyncHTTPClient, type(ret))

    def test_get_method(self):
        self.assertEqual("POST", self.actor._get_method("foobar"))
        self.assertEqual("POST", self.actor._get_method("True"))
        self.assertEqual("POST", self.actor._get_method(""))
        self.assertEqual("GET", self.actor._get_method(None))

    @testing.gen_test
    def test_generate_escaped_url(self):
        result = self.actor._generate_escaped_url("http://unittest", {"foo": "bar"})
        self.assertEqual("http://unittest?foo=bar", result)

        result = self.actor._generate_escaped_url("http://unittest", {"foo": True})
        self.assertEqual("http://unittest?foo=true", result)

        result = self.actor._generate_escaped_url(
            "http://unittest", {"foo": "bar", "xyz": "abc"}
        )
        self.assertEqual("http://unittest?foo=bar&xyz=abc", result)

        result = self.actor._generate_escaped_url(
            "http://unittest", {"foo": "bar baz", "xyz": "abc"}
        )
        self.assertEqual("http://unittest?foo=bar+baz&xyz=abc", result)

    @testing.gen_test
    def test_fetch(self):
        # Test with valid JSON
        response_dict = {"foo": "asdf"}
        response_body = json.dumps(response_dict)
        http_response = httpclient.HTTPResponse(
            httpclient.HTTPRequest("/"), code=200, buffer=io.StringIO(response_body)
        )

        with mock.patch.object(self.actor, "_get_http_client") as m:
            m.return_value = FakeHTTPClientClass()
            m.return_value.response_value = http_response

            response = yield self.actor._fetch("/")
            self.assertEqual(response_dict, response)

        # Test with completely invalid JSON
        response_body = "Something bad happened"
        http_response = httpclient.HTTPResponse(
            httpclient.HTTPRequest("/"), code=200, buffer=io.StringIO(response_body)
        )

        with mock.patch.object(self.actor, "_get_http_client") as m:
            m.return_value = FakeHTTPClientClass()
            m.return_value.response_value = http_response

            with self.assertRaises(exceptions.UnparseableResponseFromEndpoint):
                yield self.actor._fetch("/")

    @testing.gen_test
    def test_fetch_with_auth(self):
        response_dict = {"foo": "asdf"}
        response_body = json.dumps(response_dict)
        http_response = httpclient.HTTPResponse(
            httpclient.HTTPRequest("/"), code=200, buffer=io.StringIO(response_body)
        )

        with mock.patch.object(self.actor, "_get_http_client") as m:
            m.return_value = FakeHTTPClientClass()
            m.return_value.response_value = http_response

            yield self.actor._fetch("/", auth_username="foo", auth_password="bar")
            self.assertEqual(m.return_value.request.auth_username, "foo")
            self.assertEqual(m.return_value.request.auth_password, "bar")


class TestActualEnsurableBaseActor(testing.AsyncTestCase):
    def setUp(self):

        super(TestActualEnsurableBaseActor, self).setUp()
        self.actor = base.EnsurableBaseActor(
            "Unit Test Actor",
            {
                "name": "new name",
                "state": "present",
                "unmanaged": "nothing happens with this",
                "description": "Some description",
            },
        )

    @testing.gen_test
    def test_set_state(self):
        with self.assertRaises(NotImplementedError):
            yield self.actor._set_state()

    @testing.gen_test
    def test_get_state(self):
        with self.assertRaises(NotImplementedError):
            yield self.actor._get_state()
